'''
This file is part of csmlog. Python logger setup... the way I like it.
MIT License (2020) - Charles Machalow
'''

import contextlib
import datetime
import logging.handlers
import os
import pathlib
import pickle
import re
import socket
import sys
import time
import threading
import traceback
import unittest

import gspread
import gspread.auth

SCOPES = ['https://www.googleapis.com/auth/drive.file', 'https://www.googleapis.com/auth/spreadsheets']

CREDENTIALS_DIR = pathlib.Path.home()

CREDENTIALS_FILE = CREDENTIALS_DIR / '.gcreds.json'
CREDENTIALS_CACHE = CREDENTIALS_DIR / 'authorized_user.json'

DEFAULT_LOG_WORKSHEET_NAME = 'csmlog'

GOOGLE_SHEETS_MAX_CELL_CHAR_LENGTH = 50_000

LOGGER_SPREADSHEET_PREFIX = f'csmlog/{socket.gethostname()}/'

MAX_EVENTS_TO_PROCESS_PER_INTERVAL = 10_000

# soft limit to lead to a new sheet being used
MAX_EVENTS_TO_SPLIT_TO_NEW_SHEET = 100_000

MAX_OLD_LOG_SHEETS = 10

_GSPREAD = None

## Debug Code ==
_DEBUG = False
def _debug_print(s):
    if _DEBUG:
        print(s, file=sys.stderr)

def _debug_wrap(func):
    ''' a simple decorator to print function start/end time, used for debug '''
    def wrapper(*args, **kwargs):
        _debug_print(f"{datetime.datetime.now()} - About to run:  {func.__name__}")
        ret_val = func(*args, **kwargs)
        _debug_print(f"{datetime.datetime.now()} - Completed run: {func.__name__}")
        return ret_val
    return wrapper
## Debug Code ==

## Exceptions ==
class ResourceExhaustedError(RuntimeError):
    ''' raised if Google Sheets says we have talked to it too much '''
    pass

class WorkbookSpaceNeededError(RuntimeError):
    ''' raised if we need more space in the workbook '''
    pass

class WorksheetRotationNeededError(RuntimeError):
    ''' raised if we need to rotate to a new worksheet within this workbook '''
    pass
## Exceptions ==

## Misc Utils ==
@contextlib.contextmanager
def _monkeypatch(mod, name, value):
    ''' simple contextmanager for monkeypatching an object '''
    with unittest.mock.patch.object(mod, name, value):
        yield

def _natural_sort_worksheet(x):
    ''' helper to sort worksheets naturally '''
    l = re.findall(r'\d+$', x.title)
    if l:
        return int(l[0])

    return -1
## Misc Utils ==

## Resource Exhaustion Handling ==
@_debug_wrap
def _handle_resource_exhausted_error():
    ''' called whenever Google Sheets says we have exhausted our resources (API rate limiting) '''
    _debug_print("Traceback that led to resource exhaustion handling: " + traceback.format_exc())
    time.sleep(3)

def _wrap_for_resource_exhausted(func):
    ''' a decorator to auto-retry the function if it detects a resource exhaustion exception '''
    def wrapper(*args, **kwargs):
        while True:
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                if 'RESOURCE_EXHAUSTED' in str(ex):
                    _debug_print(ex)
                    _handle_resource_exhausted_error()
                    continue
                raise
    return wrapper

class _WrapperForResourceExhaustionHandling:
    '''
    A special object that takes in a gspread-based object. All calls are passed through to that object.
    When callables are given back, they will be wrapped to auto-retry on resource exhaustion
    '''
    def __init__(self, gspread):
        self._gspread = gspread

    def __repr__(self):
        return repr(self._gspread)

    def __getattribute__(self, name):
        if name != '_gspread':
            thing = getattr(self._gspread, name)
            if callable(thing):
                return _wrap_for_resource_exhausted(thing)
            return thing
        else:
            return object.__getattribute__(self, name)
## Resource Exhaustion Handling ==

def _login_and_get_gspread(credentials_file):
    ''' login and get a Sheets instance. Will prompt for login if not done before '''
    global _GSPREAD
    if not _GSPREAD:
        if not os.path.isfile(credentials_file):
            raise FileNotFoundError(f"{credentials_file} should exist before using GSheetsHandler")

        try:
            _GSPREAD = gspread.service_account(credentials_file)
            _GSPREAD._login_type = 'service_account'
        except ValueError:
            # maybe we were given oauth client id instead

            # it would be cool if we could give a custom creds path, so improvise and make it allow this.
            with _monkeypatch(gspread.auth, 'DEFAULT_CREDENTIALS_FILENAME', CREDENTIALS_FILE):
                with _monkeypatch(gspread.auth, 'DEFAULT_AUTHORIZED_USER_FILENAME', CREDENTIALS_CACHE):
                    _GSPREAD = gspread.oauth()

            _GSPREAD._login_type = 'user_oauth'

    return _GSPREAD

class GSheetsHandler(logging.StreamHandler):
    ''' Special logging handler to send events to a Google Sheet '''

    def __init__(self, logger_name, share_email=None, min_time_per_process_loop=1, max_time_per_process_loop=5, credentials_file=CREDENTIALS_FILE, start_processing_thread=True):
        ''' initializer. start_processing_thread should be True unless you don't actually want to process log events
        share_email can be an email address to share the Google Sheet workbook with. '''
        self.logger_name = logger_name
        self.gspread = _WrapperForResourceExhaustionHandling(_login_and_get_gspread(credentials_file))

        self.workbook_name = LOGGER_SPREADSHEET_PREFIX + self.logger_name
        try:
            self.workbook = _WrapperForResourceExhaustionHandling(self.gspread.open(self.workbook_name))
        except gspread.SpreadsheetNotFound:
            self.workbook = _WrapperForResourceExhaustionHandling(self.gspread.create(self.workbook_name))

        # Ensure there is a log sheet
        self._ensure_default_sheet()

        # delete sheet1
        worksheet_names = [a.title for a in self.workbook.worksheets()]
        if 'Sheet1' in worksheet_names:
            self.workbook.del_worksheet(self.workbook.worksheet('Sheet1'))

        self.share_email = share_email
        if self.share_email:
            self._make_owner_if_not_already()

        self.min_time_per_process_loop = min_time_per_process_loop
        self.max_time_per_process_loop = max_time_per_process_loop

        # rows that have not been added yet
        self._pending_rows = []
        self._pending_rows_mutex = threading.Lock()

        # keep track of the amount of time it took to add rows
        self._add_rows_time = 0

        # start processing thread
        self._start_processing_thread = start_processing_thread
        self._run_process_pending_rows = True
        self._processing_thread = threading.Thread(target=self._periodically_process_pending_rows, daemon=True)
        if self._start_processing_thread:
            self._processing_thread.start()

        logging.StreamHandler.__init__(self)

    def __repr__(self):
        return f'<GSheetsHandler {self.logger_name}>'

    def _make_owner_if_not_already(self):
        for p in self.workbook.list_permissions():
            if p['emailAddress'] == self.share_email and p['role'] == 'owner' and p['type'] == 'user':
                return

        self.workbook.share(self.share_email, perm_type='user', role='owner')

    def _ensure_default_sheet(self):
        try:
            self.sheet = _WrapperForResourceExhaustionHandling(self.workbook.worksheet(DEFAULT_LOG_WORKSHEET_NAME))
        except gspread.WorksheetNotFound:
            self.sheet = _WrapperForResourceExhaustionHandling(self.workbook.add_worksheet(DEFAULT_LOG_WORKSHEET_NAME, 1, 1))

        self.rows_in_active_sheet = self.sheet.row_count

    @_debug_wrap
    def _rotate_to_new_sheet_in_workbook(self):
        all_worksheets = sorted(self.workbook.worksheets(), key=_natural_sort_worksheet)
        all_worksheets_names = reversed([a.title for a in all_worksheets if a.title.startswith(DEFAULT_LOG_WORKSHEET_NAME)])

        def get_worksheet_by_name(name):
            for i in all_worksheets:
                if i.title == name:
                    return i

        new_sheet_list = []
        for i in all_worksheets_names:
            num_or_nothing = i.split(DEFAULT_LOG_WORKSHEET_NAME)[1]
            try:
                num = int(num_or_nothing)
            except ValueError:
                num = -1

            num = num + 1

            wks = _WrapperForResourceExhaustionHandling(get_worksheet_by_name(i))

            new_name = f'{DEFAULT_LOG_WORKSHEET_NAME}{num}'
            wks.update_title(new_name)
            new_sheet_list.append(wks)

        # remove excess old log sheets
        sheets_to_remove = list(reversed(new_sheet_list))[MAX_OLD_LOG_SHEETS:]
        for sheet in sheets_to_remove:
            _debug_print(f"Deleting sheet: {sheet}")
            self.workbook.del_worksheet(sheet)
            new_sheet_list.remove(sheet)

        self._ensure_default_sheet()

        # make the order newest to oldest
        wks_in_order = [self.sheet] + list(reversed(new_sheet_list))
        self.workbook.reorder_worksheets(wks_in_order)
        _debug_print(f"Sheet order: {wks_in_order}")

        # reset the add row time to not let this continually get run.
        #  the add_rows_time will get set by next add to the sheet
        self._add_rows_time = 0

    @_debug_wrap
    def _handle_workbook_space_needed_error(self):
        worksheets = sorted(self.workbook.worksheets(), key=_natural_sort_worksheet)
        oldest = worksheets[-1]
        _debug_print(f"Removing sheet: {oldest}")
        self.workbook.del_worksheet(oldest)

    def _add_rows_to_active_sheet(self, rows):
        ''' adds the given rows to the currently active sheet. '''
        start = time.time()
        try:
            ret = self.sheet.append_rows(rows)
            self.rows_in_active_sheet += len(rows)
            return ret
        except Exception as ex:
            # this would mean we should wait to write for a bit more.
            if 'RESOURCE_EXHAUSTED' in str(ex).upper():
                raise ResourceExhaustedError(str(ex))

            # api is down? Act the same as resource exhausted
            if 'UNAVAILABLE' in str(ex).upper():
                raise ResourceExhaustedError(str(ex))

            # this would mean we have run out of room in this sheet... try to create a new sheet/go to the next one.
            if 'ABOVE THE LIMIT' in str(ex).upper() and 'INVALID_ARGUMENT' in str(ex).upper():
                raise WorkbookSpaceNeededError(str(ex))

            raise
        finally:
            end = time.time()
            self._add_rows_time = end - start

    def _calculate_periodic_loop_sleep_time(self, time_for_process):
        ''' calculates the amount of time we should sleep based of the time it took for the
            process loop to complete '''
        if time_for_process > self.min_time_per_process_loop:
            sleep_time = 0
        else:
            sleep_time = self.min_time_per_process_loop - time_for_process

        return sleep_time

    def _periodically_process_pending_rows(self):
        ''' ran in a thread to periodically take rows and write them to sheets.
        Also may perform other actions such as rotation to keep things working smooth '''
        while self._run_process_pending_rows:
            try:
                before = time.time()
                try:
                    self.process_pending_rows()
                except ResourceExhaustedError:
                    _handle_resource_exhausted_error()
                    continue
                except WorkbookSpaceNeededError:
                    self._handle_workbook_space_needed_error()
                    continue
                except Exception as ex:
                    _debug_print(f"Exception in process_pending_rows(): {ex}")
                    continue

                if self._add_rows_time > self.max_time_per_process_loop or self.rows_in_active_sheet > MAX_EVENTS_TO_SPLIT_TO_NEW_SHEET:
                    # its taking too long to add rows to the sheet. Rotate
                    _debug_print(f"triggering rotation as the add_rows_time was: {self._add_rows_time} and rows_in_active_sheet was {self.rows_in_active_sheet}")
                    self._rotate_to_new_sheet_in_workbook()

                after = time.time()
                time.sleep(self._calculate_periodic_loop_sleep_time(after - before))
            except Exception as ex:
                _debug_print(f"Exception made it to the top of the loop in _periodically_process_pending_rows(): {traceback.format_exc()}")
                _handle_resource_exhausted_error()
                continue

    def process_pending_rows(self):
        pending_rows_copy = []
        with self._pending_rows_mutex:
            if self._pending_rows:
                # make a copy to not hold up other mutex users
                pending_rows_copy = self._pending_rows[:MAX_EVENTS_TO_PROCESS_PER_INTERVAL]

        if pending_rows_copy:
            # an exception on this means things were NOT added to the sheet
            self._add_rows_to_active_sheet(pending_rows_copy)

            # clear processed
            with self._pending_rows_mutex:
                self._pending_rows[:len(pending_rows_copy)] = []

                # if we start printing here, we are logging faster than uploading (falling behind)
                if len(self._pending_rows) > 0:
                    _debug_print(f"self._pending_rows was not empty... Size: {len(self._pending_rows)}")

    def emit(self, record):
        '''
        Called when a log record is to be sent. This just queues the events to be written to Sheets.
            Will also break up msgs that are longer than the max cell size allowed by Google.
        '''
        full_record_msg_str = str(record.msg)
        rows = [(record.asctime, record.levelname, record.pathname, record.funcName, record.lineno, full_record_msg_str),]

        if len(full_record_msg_str) > GOOGLE_SHEETS_MAX_CELL_CHAR_LENGTH:
            rows = []
            # split row into multiple
            for i in range(0, len(full_record_msg_str), GOOGLE_SHEETS_MAX_CELL_CHAR_LENGTH):
                rows.append((record.asctime, record.levelname, record.pathname, record.funcName, record.lineno, full_record_msg_str[i:i+GOOGLE_SHEETS_MAX_CELL_CHAR_LENGTH]),)

        with self._pending_rows_mutex:
            for row in rows:
                self._pending_rows.append(row)

    def flush(self):
        ''' Call to wait until all events have been reflected to sheets.
        Note: This can hang forever if we can't write to Google Sheets as fast as logging is currently happening '''
        while self._processing_thread.is_alive():
            with self._pending_rows_mutex:
                if len(self._pending_rows) == 0:
                    break
            time.sleep(.1)

    def close(self):
        ''' requests that we stop processing new pending rows '''
        self._run_process_pending_rows = False


'''
This file is part of csmlog. Python logger setup... the way I like it.
MIT License (2019) - Charles Machalow
'''

import logging
import logging.handlers
import socket

MAX_BYTES_PER_SEND = 1024

class UdpHandler(logging.StreamHandler):
    ''' handler to send live logs as raw text to a UDP socket '''
    stream = None
    def __init__(self, ip='127.0.0.1', port=5123):
        self.ip = ip
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        logging.StreamHandler.__init__(self)

    def __repr__(self):
        return '<UdpHandler %s:%s>' % (self.ip, self.port)

    def emit(self, record):
        msg = (self.format(record) + "\n").encode()
        msgLen = len(msg)
        offset = 0

        # make sure we send all
        while offset < msgLen:
            offset += self.socket.sendto(msg[offset:offset+MAX_BYTES_PER_SEND], (self.ip, self.port))



'''
This file is part of csmlog. Python logger setup... the way I like it.
MIT License (2019) - Charles Machalow
'''

import subprocess
import threading
import time

class LoggedSystemCall(object):
    ''' helpful to do a system call and log the output '''
    def __init__(self, logger):
        self.logger = logger

    def _logIo(self, process):
        ''' sends output to the logger (one line at a time) and returns the output '''
        savedOutput = ''
        while True:
            output = process.stdout.readline().decode()
            if output:
                savedOutput += output
                self.logger.debug("<CMD OUTPUT>: %s" % output.rstrip('\n'))
            else:
                break
        return savedOutput

    def check_output(self, cmd, shell=False):
        ''' similiar to subprocess.check_output but sends all output to the logger '''
        self.logger.debug("About to call: %s" % cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=shell)
        output = ''
        while process.poll() is None:
            output += self._logIo(process)
            time.sleep(.00001) # yield a bit

        self.logger.debug(".. Exit Code: %d" % process.returncode)

        if process.returncode != 0:
            raise subprocess.CalledProcessError(process.returncode, cmd, output)
        return output

    def call(self, cmd, shell=False):
        ''' similiar to subprocess.call but sends all output to the logger '''
        try:
            self.check_output(cmd, shell=shell)
            return 0
        except subprocess.CalledProcessError as ex:
            return ex.returncode




'''
This file is part of csmlog. Python logger setup... the way I like it.
MIT License (2019) - Charles Machalow
'''

import socket
import sys
import threading

MAX_UDP_PACKET_SIZE = 65535

class UdpHandlerReceiver(object):
    ''' receiver to print live logs as raw text from a UDP socket '''
    def __init__(self, ip='127.0.0.1', port=5123, bufferMaxSize=MAX_UDP_PACKET_SIZE):
        self.ip = ip
        self.port = port
        self.bufferMaxSize = bufferMaxSize


        self.__stop = False
        self.__lock = threading.Lock()
        self.__buffer = ''

    def __repr__(self):
        return '<UdpHandlerReceiver %s:%s>' % (self.ip, self.port)

    def _appendToBuffer(self, data):
        with self.__lock:
            self.__buffer += data

            # chop buffer to latest x items
            self.__buffer = self.__buffer[-self.bufferMaxSize:]

    def getBuffer(self):
        with self.__lock:
            return self.__buffer

    def requestStop(self):
        with self.__lock:
            self.__stop = True

    def shouldStop(self):
        with self.__lock:
            return self.__stop

    def recieveForever(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)
        self.socket.bind((self.ip, self.port))
        try:
            while not self.shouldStop():
                # don't block to make this easier to ctrl-c
                try:
                    data = self.socket.recv(MAX_UDP_PACKET_SIZE).decode()
                except socket.error:
                    continue

                if len(data):
                    self._appendToBuffer(data)
                    sys.stdout.write(data)

        finally:
            self.socket.close()
            del self.socket

def main():
    u = UdpHandlerReceiver()
    u.recieveForever()

if __name__ == '__main__':
    main()


from setuptools import setup
import os
import pathlib
import re
import sys

init_text = pathlib.Path(__file__).parent / 'csmlog' / '__init__.py'
version = None
for line in init_text.read_text().splitlines():
    if line.startswith('__version__'):
        version = eval(line.split('=')[1])

assert version, 'Could not find __version__!'

setup(
    name='csmlog',
    author='csm10495',
    author_email='csm10495@gmail.com',
    url='http://github.com/csm10495/csmlog',
    version=version,
    packages=['csmlog'],
    license='MIT License',
    python_requires='>=3.6',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    include_package_data = True,
    install_requires=['six', 'gspread'],
    entry_points={
        'console_scripts': [
            'csmlogudp = csmlog.udp_handler_receiver:main'
        ]
    },
)


'''
This file is part of csmlog. Python logger setup... the way I like it.
MIT License (2021) - Charles Machalow
'''

import logging
import logging.handlers
import os
import shutil
import sys
import uuid
from pathlib import Path

from csmlog.google_sheets_handler import GSheetsHandler
from csmlog.system_call import LoggedSystemCall
from csmlog.udp_handler import UdpHandler
from csmlog.udp_handler_receiver import UdpHandlerReceiver

__version__ = '0.24.0'

DEFAULT_LOG_FORMAT = '%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s'

class CSMLogger(object):
    '''
    object to wrap logging logic
    '''
    def __init__(self, appName, clearLogs=False, udpLogging=True, googleSheetShareEmail=None, formatter=None):
        self.appName = appName
        self.udpLogging = udpLogging
        self.googleSheetShareEmail = googleSheetShareEmail
        self._loggers = []

        # sets self._formatter
        self.setFormatter(formatter)

        if clearLogs:
            self.clearLogs()

        self.parentLogger = self.__getParentLogger()
        self.consoleLoggingStream = None
        self._loggers = [self.parentLogger] # keep track of all loggers

    def close(self):
        for logger in self._loggers:
            for handler in logger.handlers[:]:
                handler.close()
                logger.removeHandler(handler)

        self._loggers = []

    def getLogger(self, name):
        name = os.path.basename(name)
        loggerName = '%s.%s' % (self.appName, name) # make this a sublogger of the whole app
        logger = self.__getLoggerWithName(loggerName)
        self._loggers.append(logger)
        logger.sysCall = LoggedSystemCall(logger)
        return logger

    def __getParentLogger(self):
        logger = self.__getLoggerWithName(self.appName)
        if self.udpLogging:
            handler = UdpHandler()
            handler.setFormatter(self.getFormatter())
            logger.addHandler(handler)

        if self.googleSheetShareEmail:
            handler = GSheetsHandler(self.appName, self.googleSheetShareEmail)
            handler.setFormatter(self.getFormatter())
            logger.addHandler(handler)

        return logger

    def __getLoggerWithName(self, loggerName):
        logger = logging.getLogger(loggerName)
        logger.setLevel(1) # log all

        logFolder = self.getDefaultSaveDirectory()

        logFile = os.path.join(logFolder, loggerName + ".txt")

        formatter = self.getFormatter()

        class RotatingFileHandlerThatWillKeepWorkingOnPermissionErrorDuringRotate(logging.handlers.RotatingFileHandler):
            '''
                This class is special, it exists because on Windows, file names can't be changed while files are open.
                    So ultimately we can't rotate if 2 processes are writing to a given log file.
                    The default RotatingFileHandler will just throw the PermissionError and stop writing
                            (since it closed the stream already)
                    By throwing away the PermissionError (since logging it on every log statement would be too much),
                        and using the delay=True parameter to RotatingFileHandler, we will just continue writing to the
                        original file without rotating, via reopening the original file. If the other process closes the file,
                        it will rotate on the next log statement
            '''

            # todo: probably write test cases for rotation with multiple processes writing to one log file.
            #   ensure some sort of semi-deterministic behavior
            def rotate(self, source, dest):
                try:
                    logging.handlers.RotatingFileHandler.rotate(self, source, dest)
                except PermissionError:
                    pass

        rfh = RotatingFileHandlerThatWillKeepWorkingOnPermissionErrorDuringRotate(logFile, maxBytes=1024*1024*8, backupCount=10, delay=True)
        rfh.setFormatter(formatter)
        logger.addHandler(rfh)

        # add the log file path / folder for easy access elsewhere
        logger.logFile = logFile
        logger.logFolder = logFolder
        logger.loggerName = loggerName

        return logger

    def getFormatter(self):
        return self._formatter

    def getDefaultSaveDirectory(self):
        return self.getDefaultSaveDirectoryWithName(self.appName)

    def enableConsoleLogging(self, level=1, stream=None):
        if stream is None:
            # evaluate sys.stderr later since pytest may change it
            stream = sys.stderr

        if self.consoleLoggingStream:
            self.disableConsoleLogging()

            # recursive
            return self.enableConsoleLogging(level=level, stream=stream)
        else:
            self.consoleLoggingStream = logging.StreamHandler(stream)
            self.consoleLoggingStream.setFormatter(self.getFormatter())
            self.parentLogger.addHandler(self.consoleLoggingStream)

        self.consoleLoggingStream.setLevel(level)

    def disableConsoleLogging(self):
        if not self.consoleLoggingStream:
            raise RuntimeError("Managed console logging is not active")

        self.parentLogger.removeHandler(self.consoleLoggingStream)
        self.consoleLoggingStream = None

    def setFormatter(self, formatter=None):
        if formatter is None:
            formatter = logging.Formatter(DEFAULT_LOG_FORMAT)

        if isinstance(formatter, str):
            formatter = logging.Formatter(formatter)

        self._formatter = formatter

        for logger in self._loggers:
            for handler in logger.handlers:
                handler.setFormatter(formatter)

    @classmethod
    def getDefaultSaveDirectoryWithName(cls, appName):
        if os.name == 'nt':
            logFolder = os.path.join(os.path.expandvars("%APPDATA%"), appName)
        else:
            tmpPath = Path(f'/var/log/{uuid.uuid4()}')
            try:
                tmpPath.touch()
                tmpPath.unlink()
                tmpPath = tmpPath.parent
            except PermissionError:
                # can't use /var/log... try using ~/log/
                tmpPath = Path.home() / 'log'
                tmpPath.mkdir(exist_ok=True)

            logFolder = tmpPath / appName

        if not os.path.isdir(logFolder):
            os.makedirs(logFolder)

        return logFolder

    def clearLogs(self):
        shutil.rmtree(self.getDefaultSaveDirectory())

        # recreate empty folder
        self.getDefaultSaveDirectory()

class _CSMLoggerManager:
    ''' manages the active instance (and older instances) of CSMLogger '''

    def __init__(self):
        # loggers that are no longer default (setup() was called again, though may still be in use)
        self._oldCsmLoggers = []

        # The currently active logger
        self._activeCsmLogger = None

        # publish methods from this guy
        for name in dir(self):
            if name.startswith('_') or name.endswith('_'):
                continue

            globals()[name] = getattr(self, name)

    def getLogger(self, *args, **kwargs):
        if not self._activeCsmLogger:
            raise RuntimeError("(csmlog) setup() must be called first!")

        return self._activeCsmLogger.getLogger(*args, **kwargs)

    def close(self):
        ''' will close ALL known CSMLoggers, including active and old '''
        if not self._activeCsmLogger:
            raise RuntimeError("(csmlog) setup() must be called first!")

        self._activeCsmLogger.close()
        self._activeCsmLogger = None

        for i in self._oldCsmLoggers:
            i.close()

        self._oldCsmLoggers.clear()

    def getCSMLogger(self):
        return self._activeCsmLogger

    def enableConsoleLogging(self, *args, **kwargs):
        if not self._activeCsmLogger:
            raise RuntimeError("(csmlog) setup() must be called first!")

        return self._activeCsmLogger.enableConsoleLogging(*args, **kwargs)

    def disableConsoleLogging(self, *args, **kwargs):
        if not self._activeCsmLogger:
            raise RuntimeError("(csmlog) setup() must be called first!")

        return self._activeCsmLogger.disableConsoleLogging(*args, **kwargs)

    def setFormatter(self, formatter=None):
        if not self._activeCsmLogger:
            raise RuntimeError("(csmlog) setup() must be called first!")

        return self._activeCsmLogger.setFormatter(formatter)

    def setup(self, appName, clearLogs=False, udpLogging=True, googleSheetShareEmail=None):
        ''' must be called to setup the logger. Passes args to CSMLogger's constructor '''

        if self._activeCsmLogger is not None:
            self._activeCsmLogger.parentLogger.debug("CSMLogger was already setup. Swapping to appName: %s." % appName)
            self._oldCsmLoggers.append(self._activeCsmLogger)

        self._activeCsmLogger = CSMLogger(appName, clearLogs, udpLogging, googleSheetShareEmail)
        self._activeCsmLogger.parentLogger.debug("==== %s is starting ====" % appName)

# this will also publish all public methods to globals() for this file.
_csmLoggerManager = _CSMLoggerManager()

# legacy alias for setup()
CSMLogger.setup = setup